################################################################################
# Licensed Materials - Property of IBM Copyright IBM Corporation 2012. All Rights Reserved.
# U.S. Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP
# Schedule Contract with IBM Corp.
################################################################################

require "devops_core/util"

################################################################################
# Class Definition                                                             #
################################################################################
module DevopsLibrary
  module Util


    # This method will take an asset descriptor JSON template combine it with a set
    # of bindings to produce a complete asset configuration.
    #
    # == Params
    #
    # === Required
    # [template_file (File)]   An ERB formatted JSON template file.
    # [bindings (Hash|Object)] If a hashmap is provided, the key-value pairs will be converted into an object's 
    #                          instance variables and accessed in a template via the "@key" syntax. If an object
    #                          is provided as the value, the ERB engine will use the instance's attributes
    #                          when evaluating the template string.
    #
    # == Returns
    # [Hash]  A complete asset descriptor in the form of a hash object.  
    #
    def Util.load_library_asset_config(template_file, bindings={})

     template_str = File.read(template_file)
     asset_as_json =  DevopsCore::Util::evaluate_template(template_str, bindings)
     return DevopsCore::Util::create_hash_from_json(asset_as_json)

    end

  end # end Util
end # end DevopsLibrary
