require 'weaver/core/base.rb'

module Weaver
  module Core
    # A mixin for classes such as Node which may contain Component
    # 
    # Author:: Ed Snible
    # Copyright:: IBM
    module ComponentContainerMixin
      include Weaver::Core      # Allow us to use @@logger
      
      # Let all classes that include ComponentContainerMixin have a method to return their components
      def self.included(baseclazz)
        baseclazz.class_exec {
          # Hash of id -> Component
          def components
            # Since mixins have no initialize() to set up @components, initialize @components lazily
            @components ||= {} unless frozen?
            return @components || {}
          end
          
          # * +new_components+ Hash of id -> Component
          def components=(new_components)
            @components = new_components
          end
        }
        
        baseclazz.register_initialize_copy_handler(:initialize_copy_components)
        baseclazz.register_validate_handler(:validate_components)
        baseclazz.register_child_enumerator(:components)           
        
      end
      
      # Deeply copy @components
      def initialize_copy_components(other_component_container)
        # At this point, self.initialize_copy should have cloned @components but the values are shallow.
        @components = {}
        other_component_container.components.each do |id, component|
          dup = component.clone()
          dup.scope = self
          @components[id] = dup
        end
      end
      
      # Validate @components
      # * +out_messages+ ProblemStatus will be pushed onto this for any validation message
      # * +desired_phases+ optional Array which may be +[:predefined]:
      def validate_components(out_messages, desired_phases = nil)
        components.each { |id, component| component.validate_collect(out_messages, desired_phases) }
      end
    end
    
  end
end