require 'string'
require 'weaver/core/base'


module Weaver
  module Core
    # Implements JSON functionality for Weaver
    #
    # Author:: Ed Snible
    class Base
      def to_hash
        retval = {
          'type'         => type,
        }
        
        raise WeaverError.new("No type on #{self.fully_qualified_id} for #{root_entity.id}") if type.nil?
        
        retval['tags'] = tags.to_a unless tags.empty?
          
        children = self.amd.clone          
        self.class.child_enumerators.each do |func|
          children.merge!(send(func))
        end
        
        unique_children = children.clone
        unique_children.delete_if { |id, child| id != child.id || !child.scope.equal?(self) }
        retval.merge!(unique_children)
        
        unless unique_children.size == children.size
          aliases = children.clone
          aliases.delete_if { |id, child| id == child.id && child.scope.equal?(self) }
          retval['aliases'] = Hash[aliases.to_a.map { |id, aliaz| [id, aliaz.fully_qualified_id] }]
        end
          
        return retval
      end
      
      def to_json(*a)
        return to_hash.to_json(*a)
      end
      
      def merge_hash(h, context)
        h.each do |key, val|
          case key
          when 'type'
            @type = val.symbolize
          when 'tags'
            # @tags = Set[*val]
            @tags = Set[*(val.map { |s| s.symbolize})]
#          when 'attributes'
#            val.each do |attrname, attrval|
#              attr = amd[attrname.symbolize]
#              if attr.nil?
#                attr = Attribute.new(attrname.symbolize, self)
#                amd[attr.id] = attr
#              end
#              attr.value = attrval unless attr.unresolved_value.kind_of?(LateBinding)
##              puts "ignoring merge of attribute #{attr.fully_qualified_id} = #{attrval}" if attr.unresolved_value.kind_of?(LateBinding) 
#            end
          when 'aliases'
            val.each do |aliaz, full_id|
              # The usual case is that we are merging and the alias already exists and can be ignored
              # but if we get a new alias merge it
              if child_entities[aliaz.symbolize].nil?
                # Record the alias for later processing
                context[:aliases] << [self, full_id, aliaz.symbolize]
#                self.alias_of(root_entity.instance_eval(full_id), aliaz.symbolize)
              end
            end
          else
            child = child_entities[key.symbolize]
            if child.nil?
              if val.kind_of?(Hash)
                if val['type'] == 'attribute'
                  # Special handling for attributes, which don't have prototypes
                  child = Attribute.new(key.symbolize, self)
                  child.implicit=false
                else
                  begin
                    child = root_entity.prototype_instance(val['type'].symbolize, key.symbolize, self, nil)
                  rescue WeaverError => we
                    @@logger.warn "Cannot create #{val['type']} (#{key.symbolize.inspect}) { ... }"
                  end
                end
                
                if child.kind_of?(Attribute)
                  amd[child.id] = child
                elsif child.kind_of?(Topology) && self.is_a?(TopologyContainerMixin)
                  # Special handling of topologies because they cannot be aliased; TODO add alias support?
                  topologies[child.id] = child
                elsif child.kind_of?(User) && self.is_a?(UserContainerMixin) 
                  # Special handling of users because they cannot be aliased; TODO add alias support?
                  users[child.id] = child
                elsif child.kind_of?(Automation) && self.is_a?(AutomationContainer) 
                  # Special handling of automation because they cannot be aliased; TODO add alias support?
                  automation[child.id] = child
                elsif child.kind_of?(Component) && self.is_a?(ComponentContainerMixin) 
                  # Special handling of component because they cannot be aliased; TODO add alias support?
                  components[child.id] = child
                elsif child.kind_of?(Link) && self.is_a?(LinkContainerMixin) 
                  # Special handling of Link because they cannot be aliased; TODO add alias support?
                  links[child.id] = child
                elsif child.kind_of?(Group) && self.is_a?(GroupContainerMixin) 
                  # Special handling of Group because they cannot be aliased; TODO add alias support?
                  groups[child.id] = child
                elsif child.kind_of?(Policy) && self.is_a?(PolicyContainerMixin) 
                  # Special handling of Policys because they cannot be aliased; TODO add alias support?
                  policies[child.id] = child
                elsif child.kind_of?(LoadDriver) && self.is_a?(LoadDriverContainerMixin) 
                  # Special handling of Policys because they cannot be aliased; TODO add alias support?
                  load_drivers[child.id] = child
                elsif child.kind_of?(NetworkInterfaceController) && self.kind_of?(Node) 
                  # Special handling of NetworkInterfaceControllers because they cannot be aliased; TODO add alias support?
                  nics[child.id] = child
                elsif child.kind_of?(Provider) && self.is_a?(ProviderContainerMixin) 
                  # Special handling of Providers because they cannot be aliased; TODO add alias support?
                  providers[child.id] = child
                elsif child.kind_of?(Service) && self.is_a?(ServiceContainerMixin) 
                  # Special handling of Services because they cannot be aliased; TODO add alias support?
                  services[child.id] = child
                elsif child != nil && child.weaver_instance_of?(:storage) && self.is_a?(StorageContainerMixin)
                  # Special handling of storage because it cannot be aliased; TODO add alias support?
                  storage[child.id] = child
                elsif child.kind_of?(Artifact) && self.is_a?(ArtifactContainerMixin)
                  # Special handling of Artifacts because they cannot be aliased; TODO add alias support?
                  artifacts[child.id] = child 
                elsif !child.nil? 
                  # 'adopt' a child by using its own ID as it's 'alias'; because there is no original it will not be an alias
                  alias_of(child, child.id)
                end
                
                child.merge_hash(val, context) unless child.nil?
                 
#              else
#                puts "Skipping #{key} because #{self.fully_qualified_id} doesn't have one, it only has #{child_entities.keys.inspect}"
              end
            else
              child.merge_hash(val, context)
            end
          end
        end
         
      end
    end
    
    class Root
      def to_hash
        retval = super.to_hash
        
        retval['id'] = id
          
        # Blocks
        blocks = {}
        flatten_attributes.select { |attr| attr.unresolved_value.kind_of?(LateBinding) }.each do |attr|
          begin
            # blocks[attr.fully_qualified_id] = attr.unresolved_value.proc.to_source(:strip_enclosure => true)
            blocks[attr.fully_qualified_id] = attr.unresolved_value.to_source(:strip_enclosure => true)
            @@logger.debug "Computed block for #{attr.fully_qualified_id} as: #{blocks[attr.fully_qualified_id]}"
          rescue Sourcify::CannotParseEvalCodeError => e 
            @@logger.error "Could not call to_source on #{attr.fully_qualified_id}" 
          rescue Sourcify::MultipleMatchingProcsPerLineError => e
            @@logger.error "Could not call to_source on #{attr.fully_qualified_id}" 
          end
        end

        unless blocks.empty?
          retval['blocks'] = blocks.to_a.map {|pair| {'target'=>pair[0], 'block'=>pair[1]}}
        end
        
        #--------------------------
        # Fabio:
        # Incorporating continuous validation rules into the environment root JSON
        # -------------------------
        contval_rules = []
        @@logger.debug "Adding continuous validation rules to the environment JSON"
        self.rules.each do |id, rule|
          ## Add only continuous validation rules to the JSON
          if rule.phase == :continuous
            @@logger.debug "  rule[:#{id}]"
            rule_hash = {}
            rule_hash[:id] = rule.id
            rule_hash[:tags] = rule.tags
            rule_hash[:phase] = rule.phase
            begin
              orig_stdout = $stderr
              $stdout = StringIO.new
              rule_hash[:where] = rule.where.to_source(:strip_enclosure => true)
            rescue Sourcify::CannotParseEvalCodeError => e
              rule_hash[:where] = "Not parsed by sourcify"
            ensure
              $stdout = orig_stdout
            end 
            rule_hash[:runtime_type] = rule.contval.runtime_type
            rule_hash[:period] = rule.contval.period
            rule_hash[:command] = rule.contval.command
            rule_hash[:artifacts] = rule.contval.artifacts
            contval_rules << rule_hash
          end
        end
        
        unless contval_rules.empty?
          retval['contval_rules'] = contval_rules  
        end
        
        return retval
      end
      
      def merge_hash(h, context = {})
        context[:aliases] = [] if context[:alias].nil?
          
        h = h.clone # Make shallow copy, so we can delete without affecting caller's Hash
        if h.has_key?('id')
          @id = h.delete('id')
        end
        
        # props is an array of Hash { 'target' => <target_path>, 'source' => <source_path> }
        props = []
        if h.has_key?('propagations')
          props = h.delete('propagations')
        end

        blocks = []
        if h.has_key?('blocks')
          blocks = h.delete('blocks')
        end

        #Fabio: deal with continuous validation rules
        contval_rules = []        
        if h.has_key?('contval_rules')
          contval_rules = h.delete('contval_rules')
        end
        
        # Build up the structure        
        super
        
        # Add the aliases.  We must do this last as we need all full_ids to exist.
        context[:aliases].each do |ent, full_id, aliaz|
          actual_entity = resolve(full_id)
          if actual_entity.nil?
            @@logger.warn "Could not create alias #{ent.fully_qualified_id}.#{aliaz} for invalid entity #{full_id}"
          else
            ent.alias_of(actual_entity, aliaz)
          end 
        end
        
        # Now that containment is there set the source and target of links
        flatten_children.select { |child| child.kind_of?(Link) }.each do |link|
          link.source = instance_eval(link.source) unless link.source.nil?
          link.target = instance_eval(link.target) unless link.target.nil?
        end
        
        # Now that all entities there, set up members of groups
        flatten_children.select { |child| child.kind_of?(Group) }.each do |group|
          if group.amd.include?(:member_hack)
            group.member_hack.each do |reference|
              entity = instance_eval(reference)
              group.members[entity.id] = entity
            end
            group.amd.delete(:member_hack)
          end
        end # flatten_children.select{}.each do
        
        # Handle depends_on relationships
        flatten_children.each do |child|
          if child.respond_to?(:depends_on) and !child.depends_on.nil?
            child.depends_on.map! { |d| instance_eval(d) }
          end # if child.respond_to?(:depends_on) and ...
        end # flatten_children.each do
        
        # Now that the structure is built try to rebuild late bindings from table of propagations
        props.each do |hsh|
          source_path = hsh['source']
          target_path = hsh['target']
          target_attr = self.resolve(target_path)
          # Don't overwrite late bindings set up previously (typically hostname = late_binding { nic0.hostname })
          unless target_attr.nil? || target_attr.unresolved_value.kind_of?(LateBinding)
            code = "Proc.new { #{source_path} }"
            @@logger.debug "Creating late_binding using #{code}"
            target_attr.value = LateBinding.new(source_path, target_attr.scope, &(eval code))
          end
        end
        
        blocks.each do |hsh|
          block = hsh['block']
          target_path = hsh['target']
          target_attr = self.resolve(target_path)
          # Don't overwrite late bindings set up previously (typically hostname = late_binding { nic0.hostname })
          unless target_attr.nil? || target_attr.unresolved_value.kind_of?(LateBinding)
            code = "Proc.new { #{block} }"
            @@logger.debug "Creating late_binding using #{code}"
            target_attr.value = LateBinding.new(block, target_attr.scope, &(eval code))
          end
        end
        
        #Fabio: deal with continuous validation rules
        @@logger.debug "Converting root's JSON rules back to the model"
        contval_rules.each do |rule|
          @@logger.debug "  rule[#{rule['id']}]"
          r = Rule.new(rule['id'].to_sym)
          c = ContinuousValidation.new(r)
          r.contval = c
          self.rules[rule['id']] = r
          r.rule_id = rule['id']
          r.tags = rule['tags']
          r.phase = rule['phase'].to_sym
          r.where = eval "Proc.new { #{rule['where']} }"
          c.runtime_type = rule['runtime_type']
          c.period = rule['period']
          c.command = rule['command']
          c.artifacts = rule['artifacts'].map{|e| e.to_sym} unless rule['artifacts'].nil?
        end
        
      end
      
      def self.from_hash(h)
        retval = Weaver::Dsl::WeaverParser.new().parse(nil)
        retval.merge_hash(h)
        return retval
      end
    end
    
    class Attribute
      def to_hash
        # Ignoring the mostly-UI-specific description, default, label, group, modeled
        retval = { 'type' => type }
        begin
          unless value.nil?
            newval = value.respond_to?(:ciphertext) ? Weaver::Utils::ValueDecryptor.decrypt(value, @keyfile) : value
            retval['value'] = newval
          end 
        rescue Exception
          @@logger.warn "Could not compute #{self.fully_qualified_id}"
        end
        retval['name'] = name unless name.nil?
        retval['default'] = default unless default.nil?
        retval['required'] = false unless required
        retval['mutable'] = mutable unless !mutable
        retval['group'] = group unless group.nil?                   # Needed by IWD
        retval['label'] = label unless label.nil?                   # Needed by IWD
        retval['description'] = description unless description.nil? # Needed by IWD?
        retval['valid_values'] = valid_values unless valid_values.nil? # Needed by IWD? # TODO remove when we start serializing builtin validators
        return retval
      end
      
      def merge_hash(h, context)
        unless self.unresolved_value.kind_of?(LateBinding)
          if @from_json.nil?
            @value = h['value']
          else
            @value = @from_json.call(h['value'])
          end
        end
         
        self.default = h['default'] unless h['default'].nil?
        self.name = h['name'] unless h['name'].nil?
        self.required = h.key?('required') ? h['required'] : true
        self.mutable = h['mutable'] unless h['mutable'].nil?
        self.group = h['group'] unless h['group'].nil?
        self.label = h['label'] unless h['label'].nil?
        self.description = h['description'] unless h['description'].nil?
        self.valid_values = h['valid_values'] unless h['valid_values'].nil? || !self.valid_values.nil?
      end
      
    end
    
    class Provider
      def to_hash
        retval = super.to_hash
        retval['provider_type'] = @provider_type
        return retval
      end
      
      def merge_hash(h, context)
        h = h.clone # Make shallow copy, so we can delete without affecting caller's Hash
        if h.has_key?('provider_type')
          # Unfortunately Provider's re-use Base.type incorrectly, rename it in the JSON
          set_provider_type(h.delete('provider_type').symbolize)
        end
        
        super
      end
    end
    
    class FirewallRule
      def to_hash
        { 'ports' => @ports, 'netmask' => @netmask, 'direction' => @direction }
      end
      
#      def merge_hash(h)
#        @ports = h['ports'] unless h['ports'].nil?
#        @netmask = h['netmask'] unless h['netmask'].nil?
#        @direction = h['direction'] unless h['direction'].nil?
#      end
      
      def self.from_hash(h)
        FirewallRule.new(h['ports'], h['netmask'], h['direction'])
      end
    end
    
    class Firewall
      def to_hash
        { 'tcp' => @tcp.map { |rule| rule.to_hash }, 'udp' => @udp.map { |rule| rule.to_hash } }
      end
      
      def merge_hash(h, context)
        unless h['tcp'].nil?
          @tcp = []
          h['tcp'].each { |rule_hash| @tcp << FirewallRule.from_hash(rule_hash) }
        end
        unless h['udp'].nil?
          @udp = []
          h['udp'].each { |rule_hash| @udp << FirewallRule.from_hash(rule_hash) }
        end
      end
    end
    
    class NetworkInterfaceController
      def to_hash
        retval = super.to_hash
        retval['hostnames'] = @hostnames
        retval['firewall'] = @firewall.to_hash unless @firewall.nil?
        return retval
      end
      
      def merge_hash(h, context)
        h = h.clone # Make shallow copy, so we can delete without affecting caller's Hash
        if h.has_key?('hostnames')
          @hostnames = h.delete('hostnames')
        end
        
        if h.has_key?('firewall')
          @firewall = h.delete('firewall')
        end
        
        super
      end
    end
    
    class Link
      def to_hash
        retval = super.to_hash
        retval['source'] = @source.fully_qualified_id if @source.kind_of?(Base)
        retval['target'] = @target.fully_qualified_id if @target.kind_of?(Base)
        retval['firewall'] = @firewall.to_hash unless @firewall.nil?
        return retval
      end
      
      def merge_hash(h, context)
        h = h.clone # Make shallow copy, so we can delete without affecting caller's Hash
        if h.has_key?('source')
          source = h.delete('source')
#          @source = root_entity.instance_eval(source)
          @source = source
        end
        
        if h.has_key?('target')
          target = h.delete('target')
#          @target = root_entity.instance_eval(target)
          @target = target
        end
        
        if h.has_key?('firewall')
          @firewall = h.delete('firewall')
        end
        
        super
      end
    end
    
    class Requirement
      def to_hash
        {} # TODO
      end
    end
    
    # TODO support for cloud, realized_by, os, firewall, requirements
    class Node
      def to_hash
        retval = super.to_hash
        retval['cloud'] = @cloud.to_hash unless @cloud.nil?
        retval['os'] = @os.to_hash unless @cloud.nil?
        retval['firewall'] = @firewall.to_hash unless @firewall.nil?
          
        unless @realized_by.empty?
          retval['realized_by'] = @realized_by.select { |target| target.kind_of?(Node) }.map { |target| target.fully_qualified_id }
        end
        
        unless @requirements.empty?
          retval['requirements'] = @requirements.map { |req| req.to_hash }
        end
        
        retval.merge!(DependsOn.to_hash(self))
        
        # TODO Provisioner
        return retval
      end
      
      def merge_hash(h, context)
        h = h.clone # Make shallow copy, so we can delete without affecting caller's Hash
        if h.has_key?('cloud')
          @cloud = Cloud.new(h['type']) if @cloud.nil?
          @cloud.merge_hash(h.delete('cloud'), context)
        end
        
        if h.has_key?('os')
          @os = OperatingSystem.new if @os.nil?
          @os.merge_hash(h.delete('os'), context)
        end
        
        if h.has_key?('firewall')
          @firewall = Firewall.new if @firewall.nil?
          @firewall.merge_hash(h.delete('firewall'), context)
        end
        
        if h.has_key?('realized_by')
          realized_by_hash = h.delete('realized_by')
          # TODO
          raise WeaverError.new("Unimplemented Node.realized_by in JSON. Call JSON.generate on a folded model instead.")
        end
        
        if h.has_key?('requirements')
          reqs = h.delete('requirements')
          reqs.each do |req|
            @requirements << Requirement.new(self) # TODO we need to generate a subclass with correct attributes
          end
          @@logger.warn "Unimplemented Node.requirements in JSON"
        end

        # clone it as we take it, so that we don't overwrite late
        @depends_on = h.delete('depends_on').clone if h.has_key?('depends_on')
                
        super
      end
    end
    
    class Topology
      
      # Create a Hash representation
      def to_hash
        retval = super.to_hash
        retval['cloud'] = @cloud.to_hash unless @cloud.nil?
        # TODO Should the following be a reference to a provider instead of a deep hash?
#        retval['provisioner'] = @provisioner.to_hash unless @provisioner.nil?
        return retval
      end
      
      # Replace values with values in Hash representation
      def merge_hash(h, context)
        h = h.clone # Make shallow copy, so we can delete without affecting caller's Hash
        if h.has_key?('cloud')
          cloud_hash = h.delete('cloud')
          if @cloud.nil?
            @cloud = Cloud.new(cloud_hash['type'].symbolize)
          end
          @cloud.merge_hash(cloud_hash, context)
          
          provider = Provider.new(@cloud.type, self)
#          self.provisioner = provider
        end
        
#        if h.has_key?('provisioner')
#          provisioner_hash = h.delete('provisioner')
#          if @provisioner.nil?
#            @provisioner = Provider.new(@@@, self)
#          end
#          @provisioner.merge_hash(provisioner_hash)
#        end
        
        super
      end
    end

    class OperatingSystem
      def to_hash
        { 'type' => type, 'flavor' => flavor, 'version' => version, 'architecture' => architecture }
      end
      
      def merge_hash(h, context)
        @type = h['type']
        @flavor = h['flavor'] unless h['flavor'].nil?
        @version = h['version'] unless h['version'].nil?
        @architecture = h['architecture'] unless h['architecture'].nil?
      end
    end
        
    class Cloud
      # Create a Hash representation
      def to_hash
        retval = { 'type' => type }
        
        retval['url'] = url unless url.nil?
        retval['image'] = image unless image.nil?
        retval['description'] = description unless description.nil?
        retval['version'] = version unless version.nil?
          
        return retval
      end
      
      # Replace values with values in Hash representation
      def merge_hash(h, context)
        self.type = h['type']
        self.url = h['url'] unless h['url'].nil?
        self.image = h['image'] unless h['image'].nil?
        self.description = h['description'] unless h['description'].nil?
        self.version = h['version'] unless h['version'].nil?
      end
    end
    
    class Group
      def to_hash
        retval = super.to_hash
        retval['members'] = members.map { |id, m| m.fully_qualified_id }
        return retval
      end # def to_hash
      
      def merge_hash(h, context)
        h = h.clone # Make shallow copy, so we can delete without affecting caller's Hash
        if h.has_key?('members')
          amd[:member_hack] = Attribute.new(:member_hack, self)
          amd[:member_hack].value = h['members']
        end # fi h.has_key?('members')
        
        super
      end # def merge_hash(h,context)
    end # class Group

    #--------------------------
    # Fabio:
    # Dealing with artifacts
    # -------------------------
    class Artifact
      def to_hash
        @@logger.debug "Adding artifact to the environment JSON: #{self.inspect}"
        retval = super.to_hash
        retval['id'] = id
        retval['location'] = location unless location.nil?
        retval['source_path'] = source_path unless source_path.nil?
        retval['target_path'] = target_path unless target_path.nil?
        retval['exec'] = exec unless exec.nil?
        return retval
      end
      
      def merge_hash(h, context)
        @@logger.debug "Converting artifact's JSON back to the model: h = #{h.inspect}, context = #{context.inspect}"
        @type = h['type'].to_sym
        @location = h['location'].to_sym
        @source_path = h['source_path'] unless h['source_path'].nil?
        @target_path = h['target_path'] unless h['target_path'].nil?
        @exec = h['exec'] unless h['exec'].nil?
      end
    end
    
    class Automation
      def to_hash
        retval = super.to_hash
        retval.merge!(DependsOn.to_hash(self))
        return retval
      end # def to_hash
      
      def merge_hash(h, context)
        h = h.clone # Make shallow copy, so we can delete without affecting caller's Hash
        # clone it as we take it, so that we don't overwrite late
        @depends_on = h.delete('depends_on').clone if h.has_key?('depends_on')
        super
      end # merge_hash
    end # class Automation
    
    class DependsOn
      def self.to_hash(base)
        if base.respond_to?(:depends_on)
          unless base.depends_on.nil? or base.depends_on.empty?
            return {'depends_on'  => base.depends_on.map { |d| d.fully_qualified_id }}
          end # unless depends_on.nil? or depends_on.empty?
        end # if responds_to?(:depends_on)
        return {}
      end # def 
    end # class DependsOn
    
    
  end
end
