require 'weaver/core/base.rb'

module Weaver
  module Core
    # Represents an operating system on a Node.
    #
    # @see Os::OsType
    class OperatingSystem
      include Weaver::Core::Generation

      # Type of Operating System, such as :linux
      attr_accessor :type
      
      # Flavor/distribution of Operating System, such as :rhel
      attr_accessor :flavor
      
      # Version of Operating System, a String such as '5.7'
      attr_accessor :version
      
      # Architecture of Operating System, a Symbol such as :x86_64
      attr_accessor :architecture
      
      def initialize()
        @type = :unknown
        @flavor = :unknown
        @version = :unknown
        @architecture = :unknown
      end

      # Summarize
      def to_s
        "operating system #{@type} #{@flavor} #{@version} #{@architecture}"
      end
      
      # Validate
      # * +out_messages+ ProblemStatus will be pushed onto this for any validation message
      # * +desired_phases+ optional Array which may be <code>[:predefined]</code>
      def validate(out_messages, desired_phases = nil)
#        @@logger.debug "    TODO: validate OperatingSystem"
      end
      
      def ==(another_os)
        if another_os.nil?
          return false
        end

        retval = @type == another_os.type && @flavor == another_os.flavor &&
        @version == another_os.version && @architecture == another_os.architecture 

        return retval
      end
      
    end
  end
end