require 'weaver/core/base.rb'

module Weaver
  module Core
    # A mixin for classes such as Topology which may contain Base s with Weaver type :storage
    # 
    # Author:: Ed Snible
    # Copyright:: IBM
    module StorageContainerMixin
      include Weaver::Core      # Allow us to use @@logger
      
      # Let all classes that include StorageContainerMixin have a method to return their attachable storage
      def self.included(baseclazz)
        baseclazz.class_exec {
          # Hash of id -> Base where the Base has type :storage
          def storage
            # Since mixins have no initialize() to set up @storage, initialize @storage lazily
            @storage ||= {} unless frozen?
            return @storage || {}
          end
          
          # * +new_storage+ Hash of id -> Base
          def storage=(new_storage)
            @storage = new_storage
          end
        }
        
        baseclazz.register_initialize_copy_handler(:initialize_copy_storage)
        baseclazz.register_validate_handler(:validate_storage)
        baseclazz.register_child_enumerator(:storage)           
        
      end
      
      # Deeply copy @storage
      def initialize_copy_storage(other_storage_container)
        # At this point, self.initialize_copy should have cloned @storage but the values are shallow.
        @storage = {}
        other_storage_container.storage.each do |id, storage|
          dup = storage.clone()
          dup.scope = self
          @storage[id] = dup
        end
      end
      
      # Validate @storage
      # * +out_messages+ ProblemStatus will be pushed onto this for any validation message
      # * +desired_phases+ optional Array which may be +[:predefined]:
      def validate_storage(out_messages, desired_phases = nil)
        storage.each { |id, storage| storage.validate_collect(out_messages, desired_phases) }
      end
    end
    
  end
end
