require 'weaver/core/link'
require 'weaver/dsl/abstract_statement'

module Weaver
  module Dsl
    
    # Represents a link
    # @see Weaver::Core::Link
    #
    # Author:: Ed Snible
    # Author:: Florian Rosenberg
    # Copyright:: IBM
    class LinkStatement < AbstractStatement
      
      # * +id+ Symbol representing the entity, or a Core::Link which is the entity
      # * +parent_statement+ The enclosing AbstractStatement
      def initialize(id, parent_statement)
        @@logger.debug "link[id: #{id}]"
        super(id, Link, parent_statement)
      end

      # Summarize model
      def to_s
        "link[id: #{@id}, name: #{@name}, description: #{@description}]"
      end

      # Parse +source+ language statement
      # * +block+ ComponentStatement instance definition 
      def source(reference, &block)
        @@logger.debug "source[#{reference}]"
        
        entity = resolve_reference(reference)
        
        @model.source = entity
        if block.kind_of?(Proc)
          component = ComponentStatement.new(entity, self)
          component.instance_eval(&block)        
        end
        return @model.source        
      end

      # TODO Treat 'target' as a NodeStatement
      def target(reference, &block)
        @@logger.debug "target[#{reference}]"

        entity = resolve_reference(reference)
        
        @model.target = entity
        if block.kind_of?(Proc)
          component = ComponentStatement.new(entity, self)
          component.instance_eval(&block)        
        end
        return @model.target        
      end
      
      # reference can be a Symbol (e.g. :jke_web)
      # reference can be a Statement referring to a Weaver entity (e.g. app_server.c1)
      def resolve_reference(ref)
        return ref.model if ref.kind_of?(AbstractStatement)
        
        unless ref.kind_of?(Symbol)
          model.add_validation_failure("Link endpoint #{ref} must be a symbol or entity")
          return nil
        end
        
        retval = @model.scope.flatten_children.find { |entity| entity.id == ref }
          
        if retval.nil?
          model.add_validation_failure("Link endpoint not found #{ref}")
          return nil
        end
        
        return retval
      end
      
      # Parse +firewall+ language statement
      # * +block+ FirewallStatement instance definition 
      def firewall(&block)
        firewall_statement = FirewallStatement.new(self)
        @model.firewall = firewall_statement.model
        firewall_statement.instance_eval &block unless block.nil?
      end

    end
  end
end