require 'weaver/core/logging'
require 'weaver/dsl/requires/type_handler'
require 'weaver/core/requirements'

module Weaver
  module Dsl
    module Requires
      # Parses Weaver <code>requires :package</code> expressions, such as
      #  requires :package => "mysql-server", :version => '5+', :description => "MySQL server"
      class PackageTypeHandler < TypeHandler

        def initialize()
          super [ :package ]
        end

        def handle(parent, args)
          @@logger.debug "#{self.class} #{parent} #{args}"
          return VersionedMemberRequirement.new(parent.model, args) 
        end

      end
    end
    
    # Represents a Requirement that can be checked using set membership and has a version constraint 
    #
    # Author:: Ed Snible
    class VersionedMemberRequirement < MemberRequirement
      # * +scope+ This requirement's container, typically a Component
      # * +h+ Hash of requirement attributes
      def initialize(scope, h)
        h = h.clone # Make shallow copy, so we can delete without affecting caller's Hash
        if h.has_key?(:version)
          # We could use logic from ProductRequirement.suitable_version() version checking in a custom validate_collect
          @version = h.delete(:version)
        end
        
        super scope, h
      end
      
      # Validate this requirement
      # * +out_messages+ ProblemStatus will be pushed onto this for any validation message
      # * +desired_phases+ optional Array which may be <code>[:predefined]</code>
      def validate_collect(out_messages, desired_phases = nil)
        super
   
        unless @version.nil?
          @@logger.warn "#{@scope.fully_qualified_id} requirement #{@key}:#{@value} version checking deprecated for :package; use :product"
        end
      end
      
    end
  end
end