require 'weaver/core/logging'
require 'weaver/core/topology'

module Weaver
  module Dsl
    # Mix in TopologyContainerStatement with statements that may contain {TopologyStatement}
    #
    # Author:: Ed Snible
    # Copyright:: IBM
    module TopologyContainerStatement
      include Weaver::Core
      
      # * *Args* :
      #   -+base+ the AbstractStatement being mixed with
      def self.included(base)
        base.register_statement_handler([:topology_handler])
      end

      # * *Args* :
      #   -+sym_or_model+ <Symbol representing Weaver type, e.g. <code>:infrastructure</code> | Core::Topology
      #   -+id+ Symbol naming entity
      #   -+block+ optional block of additional metadata
      def topology_handler(sym_or_model, id, &block)
        return nil unless is_a?(Weaver::Dsl::TopologyContainerStatement)
        return nil unless sym_or_model.kind_of?(Topology) || root_statement.model.weaver_extended_type?(sym_or_model, Topology)
        return nil unless id.kind_of?(Symbol) 
        
        prototype = sym_or_model.kind_of?(Symbol) ? root_statement.model.prototype_instance(sym_or_model, id, model, weaver_stackframe) : sym_or_model
        topology(id, prototype, &block)
      end

      # If prototype is nil, we will create a core Topology, otherwise inherit attributes,
      # default attribute values, and validators from the prototype
      def topology(id, prototype = nil, &block)
        return if fail_id_redefinition(id, Topology)
        old_entity = model.child_entities[id]
        
        # in case the model responds to weaver_importable? methods check if it contains the Topology with that id
        topology_model = model.topologies[id]
#        if topology_model.nil? && root_statement.model.respond_to?(:weaver_importable?) && root_statement.model.weaver_importable?(id, Topology)
#          topology_model = root_statement.model.import_by_symbol(id)
#          model.topologies[id] = topology_model
#          root_statement.model.widen(topology_model, prototype) unless prototype.nil?
#          topology_model.scope = model
#        end
        
        if topology_model.nil?
          topology_model = prototype || root_statement.model.prototype_instance(:topology, id, model, weaver_stackframe)
          if @model.respond_to?(:provisioner) && !@model.provisioner.nil?
            @model.provisioner.mixin_provider_attributes(topology_model)
          end
          @model.topologies[id] = topology_model
        else
          root_statement.model.widen(topology_model, prototype) unless prototype.nil?
        end
        
        unless block.nil?
          begin
            topology = TopologyStatement.new(topology_model, self, :topology)
            topology.instance_eval &block
          rescue ArgumentError => ae
            # For example the statement 'component' without an ID yields
            # 'wrong number of arguments (0 for 1)'
            topology_model.add_validation_failure(" #{ae.message}", weaver_stackframe(ae.backtrace))
            @@logger.warn "Caught #{ae.inspect} processing 'topology' statement, backtrace=#{ae.backtrace}"
          rescue NameError => ne
            # For example the statement 'N1.name' yields NameError uninitialized constant
            topology_model.add_validation_failure(" #{ne.message}", weaver_stackframe(ne.backtrace))
            @@logger.warn "Caught #{ne.inspect} processing 'topology' statement, backtrace=#{ne.backtrace}"
          end
        else
          topology_model.weaver_origin = weaver_stackframe
        end

        return topology_model
      end
      
    end

  end

end