require 'rest_client'
require 'weaver/core/weaver_error'
require 'weaver/core/logging'
require 'json'

module Weaver
  module Service
    class WeaverServiceError < Weaver::Core::WeaverError
      include Weaver::Core
      
      attr_accessor :response_code
      # * +msg+ message
      # * +code+ response code
      # * +original+ chained Exception
      def initialize(msg, code=nil, original=$!)
        # set the response code and append additional information to the message for a RestClient response exception
        if (original.kind_of? RestClient::ExceptionWithResponse)
          if code.nil?
            code = original.http_code
          end
          if (!original.http_body.nil?)
            json = JSON.parse(original.http_body)
            msg << "#{rest_endpoint_error_info(json)}"
          end
        end

        # for a different type of exception, try to get its response code
        if (code.nil? && original.respond_to?("response_code"))
          code = original.response_code
        end

        super msg, original
        
        @@logger.error "#{self.to_s}"

        @response_code = code
      end

      # Returns a string with info about the response if available, or an empty string
      def rest_endpoint_error_info(response)
        str = ""
        str << "; Response code #{response['status']}" if (response['status'])
        str << "; #{response['statusText']}" if (response['statusText'])
        str << "; #{response['errorType']}:" if (response['errorType'])
        str << "; #{response['errorMessage']}" if (response['errorMessage'])
        str << "; #{response['message']}" if (response['message'])
        return(str)
      end
    end
  end
end